import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'dart:async';
import 'package:intl/intl.dart' as Intl;
import 'dart:math' as Math;
import 'package:validators/validators.dart' as Validators;

class TextUtils {
  static bool isEmpty(String text) =>
      text == null || text.isEmpty || text.trim().isEmpty;

  static bool isNotEmpty(String text) => !isEmpty(text);

  static bool equals(String a, String b) => a == b;

  ///
  ///@param timeMillis the amount of time to convert in milliseconds
  ///@param decimals amount of decimals after float point when showing days, weeks, months and years
  ///@return
  ///
  static String toHumanReadableTimeCount(int timeMillis) {
    if (timeMillis == null) return null;
    String humanReadable;
    try {
      var duration = Duration(milliseconds: timeMillis);
      humanReadable = duration.toString().split(".")[0];
    } catch (e) {
      print(e);
    }
    return humanReadable;

//    int integer;
//    double time = timeMillis?.toDouble();
//    if(time < TimeUtils.minute) {
//      return Intl.NumberFormat("00:00:#00", "en_US").format(time / TimeUtils.second);
//    }
//
//    if(time < TimeUtils.hour) {
//      int mins = (time / TimeUtils.minute).truncate();
//      int seconds = (time % TimeUtils.minute).truncate();
//      return "00:"
//          "${Intl.NumberFormat("#00", "en_US").format(mins)}:"
//          "${Intl.NumberFormat("#00", "en_US").format(seconds)}";
//    }
//    if(time < TimeUtils.day) {
//      int hours = (time / TimeUtils.hour).truncate();
//      int mins = ((time / TimeUtils.minute) % TimeUtils.minute).truncate();
//      int seconds = (time % TimeUtils.minute).truncate();
//      return "${Intl.NumberFormat("#00", "en_US").format(hours)}:"
//          "${Intl.NumberFormat("#00", "en_US").format(mins)}:"
//          "${Intl.NumberFormat("#00", "en_US").format(seconds)}";
//    }
//
//    if(time < TimeUtils.week) {
//      int days = (time / TimeUtils.hour).truncate();
//      int mins = ((time / TimeUtils.minute) % TimeUtils.minute).truncate();
//      int seconds = (time % TimeUtils.minute).truncate();
//      return "${Intl.NumberFormat("#0.00", "en_US").format(hours)}";
//    }
//    return "$time d";
  }

  ///
  ///@param bytes the amount of bytes to convert
  ///@param binarySystem true means to convert using Binary System (divide by 1024), false means
  ///          International System (divide by 1000)
  ///@param rightNotation true means to show KiB, MiB, GiB, etc if system is false. false means
  ///          to show kB, MB, GB, etc whatever value system is.
  ///@param decimals amount of decimals after float point
  ///@return
  ///
  static String toHumanReadableByteCount(int bytes,
      {bool binarySystem = true,
      bool rightNotation = false,
      int decimals = 2}) {
    int unit = binarySystem ? 1024 : 1000;
    if (bytes < unit) return "$bytes B";
    int exp = (Math.log(bytes) / Math.log(unit)).truncate();
    String pre = (binarySystem ? "kMGTPEZY" : "KMGTPEZY")[exp - 1] +
        (rightNotation ? (binarySystem ? "i" : "") : "");
    if (!rightNotation) pre = pre.toUpperCase();
    String decimalsParam = "";
    if (decimals > 0) decimalsParam = "0.";
    while (decimals-- > 0) decimalsParam += "0";
    return Intl.NumberFormat("#$decimalsParam ${pre}B", "en_US")
        .format(bytes / Math.pow(unit, exp));
  }

  ///
  ///@param distance distance in meters
  ///@param advanced true indicates distance from millimeters to kilometers, otherwise indicates distance in meters or kilometers depending on the distance; only for metric system
  ///@param metric true indicates metric system, imperial otherwise
  ///@param decimals amount of decimals after float point
  ///@return formatted distance as String
  ///
  static String toHumanReadableDistanceCount(double distance,
      {bool advanced = false, bool metric = true, int decimals = 2}) {
    String unit;
    String decimalsParam = "";
    if (decimals > 0) decimalsParam = "0.";
    while (decimals-- > 0) decimalsParam += "0";
    if (metric) {
      if (advanced) {
        int min = 10;
        if (distance < min) return "$distance mm";
        double log10(num x) {
          num log = Math.log(x) / Math.log(10) + 0.000001;
          if ((log - log.truncate()) < 0.0001) log = log.truncateToDouble();
          return log;
        }

        int exp = (log10(distance) / log10(min)).truncate();
        unit = ["cm", "dm", "m", "dam", "hm", "km"][exp - 1];
        distance /= Math.pow(min, exp);
      } else {
        if (distance < 1000)
          unit = "m";
        else {
          distance /= 1000;
          unit = "km";
        }
      }
    } else {
      if (distance < 1609.344) {
        distance *= 3.2808399;
        unit = "ft";
      } else {
        distance /= 1609.344;
        unit = "mi";
      }
    }
    return Intl.NumberFormat("#$decimalsParam $unit", "en_US").format(distance);
  }

  /// The formula is speed m/time where the speed should be in meters and time is specified in seconds base
  /// for instance:
  ///   to indicate 1000 m/seconds the call should be (1000, 1)
  ///   to indicate 1000 m/minute the call should be (1000, 60)
  ///   to indicate 1000 m/hour the call should be (1000, 3600)
  /// @param speed in meters
  /// @param time in seconds has to be 1s or 60s or 3600s; that indicates seconds, minute, hour
  /// @param metric true indicates metric system, imperial otherwise
  /// @param decimals amount of decimals after float point
  /// @return formatted speed in Km/h or Mi/h according to metric as String
  ///
  static String toHumanReadableSpeedCount(double speed, int time,
      {bool metric = true, int decimals = 2}) {
    time ??= 1;
    if (time == 1)
      time = 3600;
    else if (time == 60)
      time = 60;
    else if (time == 3600)
      time = 1;
    else
      time = 1;

    speed = speed * time.toDouble() / (metric ? 1000 : 1609.344).toDouble();
    String decimalsParam = "";
    if (decimals > 0) decimalsParam = "0.";
    while (decimals-- > 0) decimalsParam += "0";
    String unit = metric ? "km" : "mi";
    return Intl.NumberFormat("#$decimalsParam $unit/h", "en_US").format(speed);
  }
}

typedef bool Validation({AdeptController controller});

class ValidationController {
  final Validation isValid;
  final String message;

  ValidationController({this.isValid, this.message});
}

class EmptyValidationController extends ValidationController {
  EmptyValidationController({String message = "This field is required"})
      : super(
          message: message,
          isValid: ({controller}) =>
              TextUtils.isNotEmpty(controller?.rawValue?.toString()),
        );
}

class LengthValidationController extends ValidationController {
  LengthValidationController(
      {int minLength, int maxLength, String message = "Invalid length"})
      : super(
          message: message,
          isValid: ({controller}) => Validators.isLength(
              controller?.rawValue?.toString(), minLength, maxLength),
        );
}

class RepeatPasswordValidationController extends ValidationController {
  RepeatPasswordValidationController(
      {AdeptController passwordController,
      String message = "Passwords doesn't match"})
      : super(
          message: message,
          isValid: ({controller}) =>
              controller?.rawValue?.toString() ==
              passwordController?.rawValue?.toString(),
        );
}

class EmailValidationController extends ValidationController {
  EmailValidationController({String message = "Invalid email address"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isEmail(controller?.rawValue?.toString()),
        );
}

class UrlValidationController extends ValidationController {
  UrlValidationController({String message = "Invalid url"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isURL(controller?.rawValue?.toString()),
        );
}

class IPValidationController extends ValidationController {
  IPValidationController({String message = "Invalid ip address"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isIP(controller?.rawValue?.toString()),
        );
}

class AlphanumericValidationController extends ValidationController {
  AlphanumericValidationController({String message = "Invalid alphanumeric"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isAlphanumeric(controller?.rawValue?.toString()),
        );
}

class CreditCardValidationController extends ValidationController {
  CreditCardValidationController({String message = "Invalid credit card"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isCreditCard(controller?.rawValue?.toString()),
        );
}

class DateValidationController extends ValidationController {
  DateValidationController({String message = "Invalid date"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isDate(controller?.rawValue?.toString()),
        );
}

class AsciiValidationController extends ValidationController {
  AsciiValidationController({String message = "Invalid Ascii"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isAscii(controller?.rawValue?.toString()),
        );
}

class Base64ValidationController extends ValidationController {
  Base64ValidationController({String message = "Invalid Base64"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isBase64(controller?.rawValue?.toString()),
        );
}

class OnlyTextValidationController extends ValidationController {
  OnlyTextValidationController(
      {String message = "Text should contain only text"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isAlpha(controller?.rawValue?.toString()),
        );
}

class DNSValidationController extends ValidationController {
  DNSValidationController(
      {bool requireTld = true,
      bool allowUnderscores = false,
      String message = "Invalid DNS"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isFQDN(controller?.rawValue?.toString()),
        );
}

class HexadecimalValidationController extends ValidationController {
  HexadecimalValidationController({String message = "Invalid hexadecimal"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isHexadecimal(controller?.rawValue?.toString()),
        );
}

class HexColorValidationController extends ValidationController {
  HexColorValidationController({String message = "Invalid Hexcolor"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isHexColor(controller?.rawValue?.toString()),
        );
}

class ISBNValidationController extends ValidationController {
  ISBNValidationController({dynamic version, String message = "Invalid ISBN"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isISBN(controller?.rawValue?.toString(), version),
        );
}

class JsonValidationController extends ValidationController {
  JsonValidationController({String message = "Invalid Json"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isJSON(controller?.rawValue?.toString()),
        );
}

class LowercaseValidationController extends ValidationController {
  LowercaseValidationController({String message = "Invalid lowercase"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isLowercase(controller?.rawValue?.toString()),
        );
}

class UppercaseValidationController extends ValidationController {
  UppercaseValidationController({String message = "Invalid Uppercase"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isUppercase(controller?.rawValue?.toString()),
        );
}

class MongoIdValidationController extends ValidationController {
  MongoIdValidationController({String message = "Invalid MongoId"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isMongoId(controller?.rawValue?.toString()),
        );
}

class SurrogatePairValidationController extends ValidationController {
  SurrogatePairValidationController({String message = "Invalid SurrogatePair"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isSurrogatePair(controller?.rawValue?.toString()),
        );
}

class UUIDValidationController extends ValidationController {
  UUIDValidationController({dynamic version, String message = "Invalid UUID"})
      : super(
          message: message,
          isValid: ({controller}) =>
              Validators.isUUID(controller?.rawValue?.toString(), version),
        );
}

class NumericValidationController extends ValidationController {
  NumericValidationController({String message = "Text must be a number"})
      : super(
          message: message,
          isValid: ({controller}) =>
              num.tryParse((controller?.rawValue?.toString())) != null,
        );
}

class IntegerValidationController extends ValidationController {
  IntegerValidationController(
      {String message = "Text must be an integer number"})
      : super(
          message: message,
          isValid: ({controller}) =>
              int.tryParse((controller?.rawValue?.toString())) != null,
        );
}

class DecimalValidationController extends ValidationController {
  DecimalValidationController(
      {String message = "Text must be an decimal number"})
      : super(
          message: message,
          isValid: ({controller}) =>
              double.tryParse((controller?.rawValue?.toString())) != null,
        );
}

abstract class AdeptController<T> {
  List<ValidationController> validations;
  FocusNode focusNode;
  String error;

  bool validate({bool notify = true}) {
    clearError(notify: notify);
    if (validations == null || validations.isEmpty) return true;
    for (ValidationController validation in validations)
      if (!(validation?.isValid(controller: this) ?? true)) {
        setError(validation.message, notify: notify);
        return false;
      }
    return true;
  }

  bool get isValid {
    if (validations == null || validations.isEmpty) return true;
    for (ValidationController validation in validations)
      if (!(validation?.isValid(controller: this) ?? true)) return false;
    return true;
  }

  String get validationMessage {
    if (validations == null || validations.isEmpty) return null;
    for (ValidationController validation in validations)
      if (!(validation?.isValid(controller: this) ?? true))
        return validation.message;
    return null;
  }

  T get rawValue;

  dynamic get controllerValue => rawValue;

  bool get hasError => error?.isNotEmpty ?? false;

  bool get isEmpty;

  bool get isNotEmpty => !isEmpty;

  void setError(String error, {bool notify});

  void clearError({bool notify});

  void clear();

  void _feedback() async {
    for (int i = 0; i < 3; ++i)
      await Future.delayed(
          Duration(milliseconds: 100), () => HapticFeedback.vibrate());
  }
}

class AdeptTextEditingController extends TextEditingController
    with AdeptController<String> {

  AdeptTextEditingController({
    String text,
    FocusNode focusNode,
    List<ValidationController> validations,
  }) : super(text: text) {
    this.focusNode = focusNode;
    this.validations = validations;
  }

  @override
  get rawValue => text;

  @override
  get controllerValue => value;

  @override
  bool get isEmpty => text == null || text.isEmpty;

  @override
  void setError(String error, {bool notify = true}) {
    this.error = error;
    if (notify) {
      if (hasError) _feedback();
      notifyListeners();
    }
  }

  @override
  void clearError({bool notify = false}) => setError(null, notify: notify);

  @override
  void clear() {
    clearError();
    super.clear();
  }
}

class AdeptValueController<T> extends ValueNotifier<T> with AdeptController<T> {

  AdeptValueController({
    T value,
    FocusNode focusNode,
    List<ValidationController> validations,
  }) : super(value) {
    this.focusNode = focusNode;
    this.validations = validations;
  }

  @override
  get rawValue => value;

  @override
  bool get isEmpty => value == null;

  @override
  void setError(String error, {bool notify = true}) {
    this.error = error;
    if (notify) {
      if (hasError) _feedback();
      notifyListeners();
    }
  }

  @override
  void clearError({bool notify = false}) => setError(null, notify: notify);

  @override
  void clear() {
    clearError();
    value = null;
  }
}
